#ifndef _OME_ARRAY_H
//! \cond
#define _OME_ARRAY_H "$Id: OMEarray.h 312 2020-03-21 21:25:35Z geoff $"
//! \endcond

/*! \file
 * \brief OME sparse array implementation.
 */

/** \addtogroup OMEtypes
 * @{
 */

#include <map>

#include <OMErefCount.h>

typedef class OMEbaseType *OMEbaseTypeP;

// OMEarrayStorage is used internally by OMEarray, thus all of its
// members are private and OMEarray is listed as a friend.

class OMEarrayStorage final : public OMEreferenceCount
{
    friend class OMEarray;
    friend class OMEreferenceToData<OMEarrayStorage>;
public:
    typedef uint64_t ARRAY_SUBSCRIPT_t;

private:
    typedef _STD map<ARRAY_SUBSCRIPT_t, OMEbaseTypeP> SparseArrayType_t;
    SparseArrayType_t sparseArray;

    // default initializer
    OMEarrayStorage();

    // destructor
    ~OMEarrayStorage();

    virtual OMEreferenceCount *deepCopy() const override;

    bool operator==(const OMEarrayStorage &arg) const;

    bool operator!=(const OMEarrayStorage &arg) const;

    OMEbaseType &operator[](const ARRAY_SUBSCRIPT_t);

    const OMEbaseType &operator[](const ARRAY_SUBSCRIPT_t) const;

    bool deleteIndex(const ARRAY_SUBSCRIPT_t);

    bool indexExists(const ARRAY_SUBSCRIPT_t) const;

    ARRAY_SUBSCRIPT_t nextIndex(const ARRAY_SUBSCRIPT_t currentSubscript) const;

    ARRAY_SUBSCRIPT_t priorIndex(const ARRAY_SUBSCRIPT_t currentSubscript) const;

    inline uint_fast32_t elementCount() const OME_ALWAYS_INLINE
    {
        return (sparseArray.size());
    }

    inline bool isEmpty() const OME_ALWAYS_INLINE
    {
        return (sparseArray.empty());
    }

    template <typename STREAMTYPE>
    STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0,
                               uint8_t includeTypePrefix = OME_DEFAULT_COMPLEX_OUTPUT_MODE) const;
}; // end class OMEarrayStorage



class OME_DLL_EXPORT OMEarray final : public OMEreferenceToData<OMEarrayStorage>
{
private:
    //    typedef OMEarrayStorage::SparseArrayType_t SparseArrayType_t;

    explicit OMEarray(OMEarrayStorage *s)
    {
        // only used by deepCopy()
        data = s;
        // TODO:  may add unwanted reference since storage objects have 
        // initial reference count of 1
        // addReadOnlyReference();
    }

public:
    typedef OMEarrayStorage::ARRAY_SUBSCRIPT_t ARRAY_SUBSCRIPT_t;

    // default initializer
    OMEarray()
    {
        data = new OMEarrayStorage();
    }

    // copy initializer
    explicit OMEarray(const OMEarray &org) : OMEreferenceToData<OMEarrayStorage>(org.data)
    {
//        data = org.data;
//        addReadOnlyReference();
    }

    // assignment
    OMEarray &operator=(const OMEarray &arg)
    {
        if (this->data == arg.data) {
            //std::cout << "OMEarray assignment to self, ref=" << totalReferences() << "\n";
            return (*this); // asignment to self...
        }
        arg.data->addReadOnlyReference();
        dropReference(arg.data);
        return (*this);
    }

    ~OMEarray() {}

    // COMPARISON
    inline bool operator==(const OMEarray &arg) const
    {
        if (data == arg.data) {
            return (true);    // fast path
        }
        return (*data == *arg.data);
    }

    inline bool operator!=(const OMEarray &arg) const
    {
        if (data == arg.data) {
            return (false);    // fast path
        }
        return (*data != *arg.data);
    }

    inline OMEbaseType &operator[](const uint32_t i) OME_ALWAYS_INLINE
    {
        getUniqueReference();
        return ((*data)[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline OMEbaseType &operator[](const int32_t i) OME_ALWAYS_INLINE
    {
        getUniqueReference();
        return ((*data)[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline OMEbaseType &operator[](const uint64_t i) OME_ALWAYS_INLINE
    {
        getUniqueReference();
        return ((*data)[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline OMEbaseType &operator[](const uint16_t i) OME_ALWAYS_INLINE
    {
        getUniqueReference();
        return ((*data)[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline const OMEbaseType &operator[](const uint32_t i) const OME_ALWAYS_INLINE
    {
        const OMEarrayStorage &a = *data;
        return (a[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline const OMEbaseType &operator[](const int32_t i) const OME_ALWAYS_INLINE
    {
        const OMEarrayStorage &a = *data;
        return (a[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline const OMEbaseType &operator[](const uint64_t i) const OME_ALWAYS_INLINE
    {
        const OMEarrayStorage &a = *data;
        return (a[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline const OMEbaseType &operator[](const uint16_t i) const OME_ALWAYS_INLINE
    {
        const OMEarrayStorage &a = *data;
        return (a[static_cast<ARRAY_SUBSCRIPT_t>(i)]);
    }

    inline bool deleteIndex(const uint32_t i)
    {
        getUniqueReference();
        return (data->deleteIndex(static_cast<ARRAY_SUBSCRIPT_t>(i)));
    }

    inline bool deleteIndex(const uint64_t i)
    {
        getUniqueReference();
        return (data->deleteIndex(static_cast<ARRAY_SUBSCRIPT_t>(i)));
    }

    inline bool forceDeleteIndex(const ARRAY_SUBSCRIPT_t i)
    {
        // don't do getUniqueReference()
        return (data->deleteIndex(i));
    }

    inline bool indexExists(const uint32_t i) const
    {
        return (data->indexExists(static_cast<ARRAY_SUBSCRIPT_t>(i)));
    }

    inline bool indexExists(const int32_t i) const
    {
        return (data->indexExists(static_cast<ARRAY_SUBSCRIPT_t>(i)));
    }

    inline bool indexExists(const uint64_t i) const
    {
        return (data->indexExists(static_cast<ARRAY_SUBSCRIPT_t>(i)));
    }

    inline ARRAY_SUBSCRIPT_t nextIndex(const uint32_t currentSubscript) const
    {
        return (data->nextIndex(static_cast<ARRAY_SUBSCRIPT_t>(currentSubscript)));
    }

    inline ARRAY_SUBSCRIPT_t nextIndex(const int32_t currentSubscript) const
    {
        return (data->nextIndex(static_cast<ARRAY_SUBSCRIPT_t>(currentSubscript)));
    }

    inline ARRAY_SUBSCRIPT_t nextIndex(const uint64_t currentSubscript) const
    {
        return (data->nextIndex(static_cast<ARRAY_SUBSCRIPT_t>(currentSubscript)));
    }

    inline ARRAY_SUBSCRIPT_t priorIndex(const uint32_t currentSubscript) const
    {
        return (data->priorIndex(static_cast<ARRAY_SUBSCRIPT_t>(currentSubscript)));
    }

    inline ARRAY_SUBSCRIPT_t priorIndex(const int32_t currentSubscript) const
    {
        return (data->priorIndex(static_cast<ARRAY_SUBSCRIPT_t>(currentSubscript)));
    }

    inline ARRAY_SUBSCRIPT_t priorIndex(const uint64_t currentSubscript) const
    {
        return (data->priorIndex(static_cast<ARRAY_SUBSCRIPT_t>(currentSubscript)));
    }

    inline uint_fast32_t elementCount() const OME_ALWAYS_INLINE
    {
        return (data->elementCount());
    }

    inline bool isEmpty() const OME_ALWAYS_INLINE
    {
        return (data->isEmpty());
    }

    OMEarray *deepCopy() const
    {
        OMEarrayStorage *storageCopy = static_cast<OMEarrayStorage *>(data->deepCopy());
        OMEarray *copy = new OMEarray(storageCopy);
        return (copy);
    }

    /*! \brief Output an OMEarray to an output stream.
    * \param outputStream is a reference to an output stream.
    * \param indent specifies the number of spaces to indent before
    * beginning output.  If negative, no indent is performed
    * for the first line, but subsequent lines are indented
    * based on the absolute value of \a indent.
    * \param includeTypePrefix is a mask of flags indicating if type names
    * should prefix data elements, if strings should be enclosed in quotes,
    * etc.
     */
    template <typename STREAMTYPE>
    STREAMTYPE &outputOnStream(STREAMTYPE &outputStream,
            int_fast16_t indent = 0,
            uint8_t includeTypePrefix = OME_DEFAULT_COMPLEX_OUTPUT_MODE) const
    {
        data->outputOnStream(outputStream, indent, includeTypePrefix);
        return (outputStream);
    }
}; // end class OMEarray

/*! \brief Write an OMEarray to an output stream.
 */
template <typename STREAMTYPE>
inline STREAMTYPE &operator<<(STREAMTYPE &outputStream, const OMEarray &data)
{
    data.outputOnStream(outputStream, 0);
    return (outputStream);
}

#define OME_ARRAY_SUBSCRIPT_t(v) static_cast<OMEarray::ARRAY_SUBSCRIPT_t>(v)

/** @}
 */

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
