#ifndef _OME_FIXED_H
//! \cond
#define _OME_FIXED_H "$Id: OMEfixed.h 312 2020-03-21 21:25:35Z geoff $"
//! \endcond

#include <iostream> /* for debugging */
#    include <OMErefCount.h>

/*! \file
 * \brief OME fixed point number implementation.
 */

/** \addtogroup OMEtypes
 * @{
 */

typedef class OMEbaseType *OMEbaseTypeP;

#    ifndef MAPM_LIB_VERSION
typedef struct M_APM_struct *M_APM;
#    endif

/*! \brief Internal data type to generate fixed-point constant
 * from floating-point value.
 */
struct _OMEfixedConstant {
    double d;
};

typedef struct _OMEfixedConstant OMEfixedConstant;

/*! \brief Reference-counted storage for OMEfixedStorage type.
 */
class OMEfixedStorage : public OMEreferenceCount
{
    friend class OMEfixed;

private:
    M_APM numberData;

    explicit OMEfixedStorage(M_APM newVal);

public:
    OMEfixedStorage();

    explicit OMEfixedStorage(const int32_t);

    explicit OMEfixedStorage(const int64_t);

    explicit OMEfixedStorage(const double);

    explicit OMEfixedStorage(const char *);

    ~OMEfixedStorage();

    virtual OMEreferenceCount *deepCopy() const override;

    bool operator==(const OMEfixedStorage &arg) const;

    bool operator!=(const OMEfixedStorage &arg) const;

    bool operator<=(const OMEfixedStorage &arg) const;

    bool operator<(const OMEfixedStorage &arg) const;

    bool operator>=(const OMEfixedStorage &arg) const;

    bool operator>(const OMEfixedStorage &arg) const;

    explicit operator char *() const; // not thread safe

    char *asStringText(char *bfr, uint_fast16_t bfrLen) const;

    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0) const
    {
        char outString[256];

        outputStream << asStringText(outString, sizeof(outString));
        return (outputStream);
    }
}; // end class OMEfixedStorage

/*! \brief Public interface to OME fixed-point type.
 */
class OME_DLL_EXPORT OMEfixed : public OMEreferenceToData<OMEfixedStorage>
{
private:
    OMEfixed(OMEfixedStorage *f)
    {
        data = f;
    }

public:
    OMEfixed()
    {
        data = new OMEfixedStorage();
    }

    OMEfixed(const OMEfixed &org)
    {
        data = org.data;
        addReadOnlyReference();
    }

    explicit OMEfixed(const int32_t i)
    {
        data = new OMEfixedStorage(i);
    }

    explicit OMEfixed(const uint32_t ui)
    {
        data = new OMEfixedStorage((int32_t) ui);
    }

    explicit OMEfixed(const int64_t i64)
    {
        data = new OMEfixedStorage(i64);
    }

    explicit OMEfixed(const uint64_t ui64)
    {
        data = new OMEfixedStorage((int64_t) ui64);
    }

    explicit OMEfixed(const double d)
    {
        data = new OMEfixedStorage(d);
    }

    explicit OMEfixed(M_APM newVal)
    {
        data = new OMEfixedStorage(newVal);
    }

    explicit OMEfixed(const char *numberText)
    {
        data = new OMEfixedStorage(numberText);
    }

    ~OMEfixed() {}

    // ASSIGNMENT
    OMEfixed &operator=(const OMEfixed &arg)
    {
        if (this->data == arg.data) {
            std::cerr << "OMEfixed assignment to self, ref=" << totalReferences() << "\n";
            return (*this); // asignment to self...
        }
        arg.data->addReadOnlyReference();
        dropReference();
        data = arg.data;
        return (*this);
    }

    OMEfixed &operator=(const int32_t i)
    {
        dropReference();
        data = new OMEfixedStorage(i);
        return (*this);
    }

    OMEfixed &operator=(const uint32_t ui)
    {
        dropReference();
        data = new OMEfixedStorage((int32_t) ui);
        return (*this);
    }

    OMEfixed &operator=(const double d)
    {
        dropReference();
        data = new OMEfixedStorage(d);
        return (*this);
    }

    // COMPARISON
    bool operator==(const OMEfixed &arg) const
    {
        if (data == arg.data) {
            return (true);    // fast path
        }
        return (*data == *arg.data);
    }

    bool operator!=(const OMEfixed &arg) const
    {
        if (data == arg.data) {
            return (false);    // fast path
        }
        return (*data != *arg.data);
    }

    bool operator<(const OMEfixed &arg) const
    {
        return (*data < *arg.data);
    }

    bool operator>(const OMEfixed &arg) const
    {
        return (*data > *arg.data);
    }

    bool operator<=(const OMEfixed &arg) const
    {
        return (*data <= *arg.data);
    }

    bool operator>=(const OMEfixed &arg) const
    {
        return (*data >= *arg.data);
    }

    OMEfixed *deepCopy() const
    {
        OMEfixedStorage *storageCopy = static_cast<OMEfixedStorage *>(data->deepCopy());
        OMEfixed *copy = new OMEfixed(storageCopy);
        return (copy);
    }

    OMEfixed &operator+=(const OMEfixed &arg);

    OMEfixed &operator+=(const double arg);

    OMEfixed &operator-=(const OMEfixed &arg);

    OMEfixed &operator-=(const double arg);

    OMEfixed &operator*=(const OMEfixed &arg);

    OMEfixed &operator*=(const double arg);

    OMEfixed &operator/=(const OMEfixed &arg);

    OMEfixed &operator/=(const double arg);

    OMEfixed operator+(const OMEfixed &arg) const;
    OMEfixed operator-(const OMEfixed &arg) const;
    OMEfixed operator*(const OMEfixed &arg) const;
    OMEfixed operator/(const OMEfixed &arg) const;

    explicit operator double() const;

    explicit operator float() const;

    explicit operator int32_t() const;

    explicit operator int64_t() const;

    explicit operator uint64_t() const;

    explicit operator uint32_t() const;

    explicit operator uint16_t() const;

    explicit operator uint8_t() const;




    template <typename STREAMTYPE> STREAMTYPE &outputOnStream(STREAMTYPE &outputStream, int_fast16_t indent = 0) const
    {
        data->outputOnStream(outputStream, indent);
        return (outputStream);
    }
}; // end class OMEfixed

template <typename STREAMTYPE>
inline STREAMTYPE &operator<<(STREAMTYPE &outputStream, const OMEfixed &data)
{
    data.outputOnStream(outputStream, 0);
    return (outputStream);
}

/** @}
 */

#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
