#ifndef _OME_OBJECT_H
//! \cond
#define _OME_OBJECT_H "$Id: OMEobject.h 452 2020-07-23 20:19:23Z geoff $"
//! \endcond

#include <deque>

#include <OMEtype.h>
#include <OMEclass.h>
#include <utils/base/atomic_values.h>

/** \defgroup OMEobject OME Object Classes
 * @{
 */

typedef class OMEthread *OMEthreadP;


/*! \brief Base class and abstract interface for an object within a
 * FARGOS/VISTA Object Management Environment.
 */
class OMEobject : public OMEreferenceCount {
friend class OMEoidStorageInternal;
private:
	OME_FAST_CALL void addOIDref(OMEoidStorageInternal *ref);
	OME_FAST_CALL void removeOIDref(OMEoidStorageInternal *ref);

	OMEoid			    *defaultOID;
	mutable uint64_t	totalTicks;
	mutable uint32_t    totalSlices;
protected:
	mutable uint32_t	totalInvocations;
	uint32_t			instanceID;
	OMEentityID		    *entityID;
	OMEoidStorageInternal	*oidReferences;
public:
	const OMEclass		*classInfo;	//!< Pointer to class description
	OMEoid			    *metaObject; //!< Pointer to meta-object

protected:
	// allocate, destroy storage for an object
	explicit OMEobject(OMEclass *classData);

	virtual ~OMEobject();
public:

    /*! \brief Returns a Boolean indicating if the object is
     * representing a thread.
     */
	virtual bool isThread() const { return (false); }

	OME_FAST_CALL void getOIDdata(OMEoidID *idInfo) const;

    //! \brief Returns read-only reference to object's default object Id.
	const OMEoid &getDefaultOID();

    /*! \brief Attach an meta object to the object.
     */
	OME_FAST_CALL void setMetaObject(const OMEoid *obj) {
		delete metaObject;
        metaObject = (obj == nullptr) ? nullptr : new OMEoid(*obj);
	}

    /*! \brief Invoke method against an object.
     */
	OME_FAST_CALL virtual OMEthread *invokeMethod(int *errRet,
		const OMEstring &methodName,
		const uint32_t hashIndx, const OMEoid &objRefData,
		const uint32_t argc, const OMEarray &argv, const uint32_t signature,
		const OMEtype *fromObject = nullptr,
        const OMEthread *fromThread = nullptr,
		const OMEoid *notify = nullptr,
		const int inheritanceLevel = -1, const int mustBeAtLevel = -1) = 0;

    /*! \brief Perform RPC-style method invocation against an object.
     */
	OME_FAST_CALL OMEthread *invokeMethodAndWaitForReply(int *errRet,
		const OMEstring &methodName,
		const uint32_t hashIndx, const OMEoid &objRefData,
		const uint32_t argc, const OMEarray &argv, const uint32_t signature,
		OMEthread *fromThread,	// thread gets return result, sleeps
		const int inheritanceLevel = -1);

    /*! \brief Add tick counts to per-object statistics
     */
	OME_FAST_CALL void addTickCount(const int64_t tickCount) const OME_ALWAYS_INLINE {
		atomicIncrement_uint64(&totalTicks, tickCount);
		atomicIncrement_uint32(&totalSlices, 1);
	}

    /*! \brief Return number of elapsed time in ticks spent executing object's
     * methods
     */
	OME_FAST_CALL int64_t getTickCount() const OME_ALWAYS_INLINE {
        return (totalTicks);
    }

    //! \brief Return number of time slices allocated to object's methods.
	OME_FAST_CALL uint32_t getSliceCount() const OME_ALWAYS_INLINE {
        return (totalSlices);
    }

    //! \brief Return number of methods invoked against object.
	OME_FAST_CALL uint32_t getInvocationCount() const OME_ALWAYS_INLINE {
        return (totalInvocations);
    }
}; // end class OMEobject

/*! \brief Record to track methods that can be invoked upon an object.
 *
 * Conceptually a private class associated with OMEobjectInstance.
 */
class OMEallowEntry {
friend class OMEobjectInstance;
private:
	OMEstring	methodName;
	int		invocationLevel;
	bool		permanent;

	OMEallowEntry(const OMEstring &methodNm, const int level,
		const bool always = false) : methodName(methodNm) {
//		methodName = methodNm;
		invocationLevel = level;
		permanent = always;
	}
public:
	~OMEallowEntry()	{}
}; // end class OMEallowEntry

typedef OMEallowEntry *OMEallowEntryP;

/*! \brief Subclass of OMEobject that provides access to
 * a local object.
 */
class OMEobjectInstance final : public OMEobject {
public:
	// final method executed in tear-down of object
	static int Object_delete(OMEthread *);
private:
	void				**instanceData;
	_STD deque<OMEallowEntryP>	allowedMethods;
	_STD deque<OMEthreadP>		queuedThreads;
	uint32_t				activeThreadCount;
	uint32_t				totalThreadCount;

	// not directly thread-safe, but private so they are only
	// called by member functions that are thread safe.
	OME_FAST_CALL bool threadAllowed(OMEthread *thread, const OMEallowEntry *entry = 0);

	OMEthread *unblockNextThread();

	OME_FAST_CALL OMEthread *unblockQueuedThread(const OMEallowEntry *entry = nullptr);
protected:
	// allocate, destroy storage for an object
	OMEobjectInstance(OMEclass *classData);

	~OMEobjectInstance();
public:
	static void defineOMEobjectClass();

	// remaining are thread safe...
	OME_FAST_CALL static OMEoid *createObject(OMEclass *objClass,
		const OMEassoc &acl, const uint32_t argc, const OMEarray &argv,
		const OMEtype *fromObject = nullptr,
		const OMEthread *fromThread = nullptr, const OMEoid *notify = nullptr,
		OMEtype *errInfo = nullptr);

	OME_FAST_CALL static OMEoid *createObjectOfClass(const uint32_t argc,
		const OMEarray &argv, const OMEassoc &acl,
		const OMEstring &name, const OMEtype *fromObject = nullptr,
		const OMEthread *fromThread = nullptr, const OMEoid *notify = nullptr,
		const uint32_t verID = 0, const OMEstring *nmSpace = nullptr,
		OMEtype *errInfo = nullptr);

	OME_FAST_CALL static OMEoid *createObjectOfClass(const uint32_t argc,
		const OMEarray &argv, const OMEassoc &acl,
		const OMEstring &name,
		const OMEstring &user, const OMEstring &pw,
		const uint32_t verID = 0, const OMEstring *nmSpace = nullptr);

	OME_FAST_CALL static OMEoid *importObject(const OMEstring &data, bool duplicate,
		OMEtype *errInfo = nullptr);

	OME_FAST_CALL void deleteThisObject(const OMEoid &objRefData,
		const OMEthread *fromThread); // request deletion 

	OME_FAST_CALL void *getInstanceDataAtLevel(uint_fast16_t lvl) const OME_ALWAYS_INLINE {
		return (instanceData[lvl]);
	}

	OME_FAST_CALL const OMEinvocationData *findRoutine(int *retGraphLevel,
		const OMEstring &methodName, const uint32_t hashIndx,
		const uint32_t argc, const OMEarray &argv,
		const uint32_t signature = 0, const int inheritanceLevel = -1,
		const int mustBeAtLevel = -1);
	
	virtual OME_FAST_CALL OMEthread *invokeMethod(int *errRet, const OMEstring &methodName,
		const uint32_t hashIndx, const OMEoid &objRefData,
		const uint32_t argc, const OMEarray &argv,
		const uint32_t signature = 0,
		const OMEtype *fromObject = nullptr,
        const OMEthread *fromThread = nullptr,
		const OMEoid *notify = nullptr, const int inheritanceLevel = -1,
		const int mustBeAtLevel = -1) override;

	OME_FAST_CALL int sendMessage(const OMEstring &methodName, const OMEarray &argv,
		OMEoid &destObj);

	OME_FAST_CALL void noteThreadTerminated(class OMEthread *, bool modified,
		uint32_t exitStatus, bool wasPreemptable);

	void dropActiveThreadReference();

	OME_FAST_CALL uint_fast16_t listAllowedMethods(OMEarray &result) const;

	OME_FAST_CALL void allowMethod(const OMEthread *thread, const OMEstring &methodName,
		const OMEstring &className, const bool always = false);

	OME_FAST_CALL void allowMethod(const OMEstring &methodName,
		const OMEstring &className, const bool always = false);

	OME_FAST_CALL void allowMethod(const OMEstring &methodName, const int level = -1,
		const bool always = false);

}; // end class OMEobjectInstance


/** @}
 */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
