#ifndef _ATOMIC_VALUES_H_
//! \cond
#define _ATOMIC_VALUES_H_ "$Id: atomic_values.h 454 2020-07-23 20:22:23Z geoff $"
//! \endcond

/* Copyright (C) 2010 - 2019 FARGOS Development, LLC */

/*! \file
 * \brief Atomic operations.
 */

/** \addtogroup FARGOSbase
 *
 * @{
 */

/** \defgroup FARGOSatomic FARGOS Atomic Integer Routines
 *
 * \brief The atomic routines implement compare-and-swap and atomic
 * variable increment functions.
 *
 * @{
 */


#ifdef _MSC_VER /* compiled by Microsoft Visual Studio */

#include <winsock2.h>
#include <Windows.h>
#endif
#include <stdint.h>
#include <utils/base/compiler_hints.h>

#if __GNUC__
/*! \brief Implement an inline compare-and-swap routine for a particular type.
 */
#define DEFINE_COMPARE_AND_SWAP_TYPE(suffix, typeName) inline typeName compareAndSwap_ ## suffix (typeName *loc, typeName oldVal, typeName newVal) { \
  typeName result = __sync_val_compare_and_swap(loc, oldVal, newVal); \
  return (result); \
}

/*! \brief Implement an inline compare-and-swap routine for a distinct native
 * type, but return a Boolean indicating success.
 */ 
#define DEFINE_DID_COMPARE_AND_SWAP_TYPE(suffix, typeName) inline bool did_compareAndSwap_ ## suffix (typeName *loc, typeName oldVal, typeName newVal) { \
  bool result = __sync_bool_compare_and_swap(loc, oldVal, newVal); \
  return (result); \
}
#endif /* __GNUC__ */

/*! \brief Implement an atomic increment operator for a native type.
 */
#define DEFINE_ATOMIC_INCREMENT_TYPE(suffix, typeName) inline typeName atomicIncrement_ ## suffix(typeName *loc, int32_t amount) { typeName newVal; \
    bool ok; do { typeName cur = *(loc); newVal = cur + (amount); \
        ok = did_compareAndSwap_ ## suffix (loc, cur, newVal); \
    } while (ok == false); \
    return (newVal); \
}

#if _MSC_VER /* compiled by Microsoft Visual Studio */

inline void *compareAndSwap_Pointer(void **loc, void *oldVal, void *newVal) {
  void *orig = InterlockedCompareExchangePointer(loc, newVal, oldVal);
  return (orig);
}

inline bool did_compareAndSwap_Pointer(void **loc, void *oldVal, void *newVal) {
  void *orig = InterlockedCompareExchangePointer(loc, newVal, oldVal);
  return (orig == oldVal);
}

inline uint32_t compareAndSwap_uint32(uint32_t *loc, uint32_t oldVal, uint32_t newVal) {
  uint32_t orig = InterlockedCompareExchange(loc, newVal, oldVal);
  return (orig);
}

inline bool did_compareAndSwap_uint32(uint32_t *loc, uint32_t oldVal, uint32_t newVal) {
  uint32_t orig = InterlockedCompareExchange(loc, newVal, oldVal);
  return (orig == oldVal);
}

inline uint64_t compareAndSwap_uint64(uint64_t *loc, uint64_t oldVal, uint64_t newVal) {
  uint64_t orig = InterlockedCompareExchange64((int64_t *) loc, (int64_t) newVal, (int64_t) oldVal);
  return (orig);
}

inline bool did_compareAndSwap_uint64(uint64_t *loc, uint64_t oldVal, uint64_t newVal) {
  uint64_t orig = InterlockedCompareExchange64((int64_t *) loc, newVal, oldVal);
  return (orig == oldVal);
}

DEFINE_ATOMIC_INCREMENT_TYPE(uint32, uint32_t);
DEFINE_ATOMIC_INCREMENT_TYPE(uint64, uint64_t);

#ifdef __cplusplus
template <NUMTYPE> incrementAtomically(NUMTYPE *loc, NUMTYPE amt)
{
    NUMTYPE newVal;
    bool ok;
    do {
        NUMTYPE cur = *(loc);
        newVal = cur + (amount);
        if ((typeid(NUMTYPE) == typeid(int32_t) || (typeid(NUMTYPE) == typeid(uint32_t)) {
            NUMTYPE orig = InterlockedCompareExchange((uint32_t), newVal, cur);
            ok = orig == cur;
        } else if ((typeid(NUMTYPE) == typeid(int64_t)) || (typeid(NUMTYPE) == typeid(uint64_t))) {
            NUMTYPE orig = InterlockedCompareExchange64((uint64_t), newVal, cur);
            ok = orig == cur;
        }
    } while (ok == false);
    return (newVal);
}
#endif


#endif /* _MSC_VER */

#ifdef __GNUC__
DEFINE_COMPARE_AND_SWAP_TYPE(Pointer, void *);
DEFINE_DID_COMPARE_AND_SWAP_TYPE(Pointer, void *);

DEFINE_COMPARE_AND_SWAP_TYPE(uint32, uint32_t);
DEFINE_DID_COMPARE_AND_SWAP_TYPE(uint32, uint32_t);

DEFINE_COMPARE_AND_SWAP_TYPE(uint64, uint64_t);
DEFINE_DID_COMPARE_AND_SWAP_TYPE(uint64, uint64_t);

DEFINE_ATOMIC_INCREMENT_TYPE(uint32, uint32_t);
DEFINE_ATOMIC_INCREMENT_TYPE(uint64, uint64_t);

#if 0
inline void *compareAndSwapPointer(void **location, void *oldVal, void *newVal) {
    void *result = __sync_val_compare_and_swap(location, oldVal, newVal);
    return (result);
}
#endif

#ifdef __cplusplus
template <typename NUMTYPE> NUMTYPE incrementAtomically(NUMTYPE *loc, NUMTYPE amount)
{
    NUMTYPE newVal;
    bool ok;
    do {
        NUMTYPE cur = *(loc);
        newVal = cur + (amount);
        ok = __sync_bool_compare_and_swap(loc, cur, newVal);
    } while (ok == false);
    return (newVal);
}

/** @} */

/** @} */
#endif
#endif /* __GNUC__ */

#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
