#ifndef _BLOOM_FILTER_HPP_
//! \cond
#define _BLOOM_FILTER_HPP_ "$Id: bloom_filter.hpp 454 2020-07-23 20:22:23Z geoff $"
//! \endcond
// Copyright (C) 2012 - 2019 FARGOS Development, LLC

/*! \file
 * \brief Bloom filter implementation.
 */

#include <utils/base/compiler_hints.h>

/** \addtogroup FARGOSbase
 * @{
 */

/** \defgroup FARGOSbloomFilter Bloom Filter Presence Detector
 *
 * \brief A Bloom filter enables quickly determining if a key might
 * be present in a table.  It is most useful in situations where an
 * application might be presented with many keys that are not of interest.
 * In a trading system, these typically allow filtering out price change
 * announcements for symbols that are not of interest.
 * @{
 */

/*! \brief Templated Bloom filter.
 * \param HASHTYPE is the type of the result of the hash function.
 * \param KEYTYPE is the type of the key to be hashed.
 *
 * The Bloom filter is a probabilistic data structure that can be used
 * to quickly discard keys as not being of interest.  This is achieved
 * by testing up to 3 distinct hashes that resolve to bits within a logical
 * array of bits.  If a key to be tested does not hash to any value that
 * has a bit set, one can authoritatively conclude that the key was never
 * previously selected.
 *
 * In contrast, if one of the hashes does correspond to an already set
 * bit, one can only conclude that there is a possibility that the
 * key might be of interest.
 *
 * Typical usage is in processing subscriptions to data feeds that serve
 * a large population of potential keys for which only a small subset
 * is usually of interest.
 */
template<typename HASHTYPE, typename KEYTYPE=uint32_t> class BloomFilterForKey {
protected:
    enum {
        BIT_WIDTH = sizeof(HASHTYPE) * 8,   // assuming 8 bits per byte
        TBL_SIZE = (1 << BIT_WIDTH),
        HASH_MASK = (1 << BIT_WIDTH) - 1
    };
    uint64_t    presenceTable[TBL_SIZE / 64];

/*! \brief
 * \param v is source value 
 * \param n is width of field in bits
 * \param f is starting bit position, least significant bit is 0
 * \param t is destination bit position
 */
#define SHIFT_N_BITS_TO_POSITION(v,n,f,t) (((f >= t) ? ((v) >> (f - t)) : ((v) << (t - f))) & (((1 << n) - 1) << t))

    /*! \brief Hash 1 for Bloom filter.
     */
    HASHTYPE hash1(KEYTYPE key) const OME_CONST_FUNCTION {
        KEYTYPE k1 = key >> BIT_WIDTH;
        KEYTYPE k2 = key + k1;
        HASHTYPE result = (key + k2) & ((1 << (BIT_WIDTH - 2)) - 1);
        result |= 3 << (BIT_WIDTH - 2);
        return (result);
    }

    HASHTYPE hash2(KEYTYPE key) const OME_CONST_FUNCTION {
    /* We expect 0's in the high-order bits for normal integer keys.
     * Assuming little-endian order, we'll also expect 0's in the high-order
     * bits if the source was a string; the least significant bits will
     * correspond to the first character in string.
     */
        // bits 0-4 to bits 0-4
        KEYTYPE k1 = SHIFT_N_BITS_TO_POSITION(key, 5, 0, 0);
        // bits 12-8 to bits 9-5
        KEYTYPE k2 = SHIFT_N_BITS_TO_POSITION(key, 3, 8, 5);
        // bits 19-16 to bits 13-10
        KEYTYPE k3 = SHIFT_N_BITS_TO_POSITION(key, 4, 16, 8);
        // bits 25,24 to bits 15,14
        KEYTYPE k4 = SHIFT_N_BITS_TO_POSITION(key, 4, 24, 12);
        HASHTYPE result = ((k1 ^ k2 ^ k3 ^ k4) + key) & HASH_MASK;
        // NOTE: constant expression, so evaluated at compile-time
        if (sizeof(KEYTYPE) == 8) { // 64-bit value
            KEYTYPE k5 = SHIFT_N_BITS_TO_POSITION(key, 5, 32, 16);
            KEYTYPE k6 = SHIFT_N_BITS_TO_POSITION(key, 3, 40, 21);
            KEYTYPE k7 = SHIFT_N_BITS_TO_POSITION(key, 4, 48, 24);
            KEYTYPE k8 = SHIFT_N_BITS_TO_POSITION(key, 4, 52, 28);
            result = ((k5 ^ k6 ^ k7 ^ k8) + key) & HASH_MASK;
        }
        return (result);
    }

    HASHTYPE hash3(KEYTYPE key) const OME_CONST_FUNCTION {
        KEYTYPE k1 = key >> (BIT_WIDTH - 1);
        HASHTYPE result = (key - k1) & ((1 << (BIT_WIDTH - 2)) - 1);
        result |= 1 << (BIT_WIDTH - 1);
        return (result);
    }

    void setPresenceBit(HASHTYPE hashedVal) {
        uint32_t sub = hashedVal >> 6;  // 64 = 2^6
        OME_PREFETCH(presenceTable + sub, 1, 3); // read-write, retain if at all possible
        uint32_t bit = hashedVal & 63; // get bit
        presenceTable[sub] |= (static_cast<uint64_t>(1) << bit);
    }

    bool isPresenceBitSet(HASHTYPE hashedVal) const {
        uint32_t sub = hashedVal >> 6;  // 64 = 2^6
        OME_PREFETCH(presenceTable + sub, 0, 3); // read-only, retain if at all possible
        uint32_t bit = hashedVal & 63; // get bit
        uint64_t bitVal = presenceTable[sub] & (static_cast<uint64_t>(1) << bit);
        return ((bitVal != 0));
    }

public:
    //! \brief Construct a Bloom filter
    BloomFilterForKey() {
        memset(presenceTable, 0, sizeof(presenceTable));
    }

    ~BloomFilterForKey() {
    }

    //! \brief Clear all presence bits from the Bloom filter.
    void clear() {
        memset(presenceTable, 0, sizeof(presenceTable));
    }

    /*! \brief Add a key to the Bloom filter.
     *
     * \param key specifies the key to add to the Bloom filter
     */
    void addKey(KEYTYPE key) OME_ALWAYS_INLINE {
        HASHTYPE h1 = hash1(key);
        HASHTYPE h2 = hash2(key);
        HASHTYPE h3 = hash3(key);
        setPresenceBit(h1);
        setPresenceBit(h2);
        setPresenceBit(h3);
    }

    /*! \brief Test to see if key could be of interest.
     *
     * \param key specifies the key to test
     *
     * \retval false It is not possible for the key to have been
     * previously added to the filter.
     * \retval true It is possible, but not guaranteed, that the
     * key may have been added to the filter.
     */
    bool couldBeSelected(KEYTYPE key) const OME_ALWAYS_INLINE {
        HASHTYPE h1 = hash1(key);
        if (isPresenceBitSet(h1) == false) {
            return (false);
        }
        HASHTYPE h2 = hash2(key);
        if (isPresenceBitSet(h2) == false) {
            return (false);
        }
        HASHTYPE h3 = hash3(key);
        if (isPresenceBitSet(h3) == false) {
            return (false);
        }
        // all hashes indicate a collision
        return (true);
    }
}; // end class BloomFilterForKey

/** @} */

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
