#ifndef _FLUSH_PROFILERS_H_
//! \cond
#define _FLUSH_PROFILERS_H_ "$Id: flush_profilers.h 316 2020-03-26 16:18:40Z geoff $"
//! \endcond

#ifndef _WIN32
#include <stdio.h>
#include <unistd.h>
#include <dlfcn.h>

/*! \file
 * \brief Interfaces to flush profiling buffers.
 *
 * It is sometimes impossible to run a program to error-free completion
 * in order to record the contents of the profiling buffers.  This could be
 * due to bugs that cause the program to crash or simply because it normally
 * runs forever.
 * 
 * The special interfaces defined here allow the profiling data to be saved
 * to disk without requiring a successful run to completion.
 */

/*! \brief Flush profiling buffers - POSIX-specific
 *
 * \param forceExit indicates if the process should be terminated.
 * A value of 1 indicates _exit() should be called; a value of 2 indicates
 * exit() should be called, which invokes the atexit() handlers and gracefully
 * tears down objects created before main() was invoked.
 */
inline void flushProfilingBuffers(int forceExit) {
    typedef void (*voidFuncCallFP)();

    void *dlHandle = dlopen(0, RTLD_LAZY);
    /* Based on implementation in gcc-V.N.x/libgcc/config/arc/gmon/gcmon.c:
     * _mcleanup() would be the normal routine, but it disables profiling,
     * writes the data and frees buffers.
     * __write_profiling() just writes current image if profiling was
     * enabled and doesn't change any state.
     */
    void *monitorCleanup = dlsym(dlHandle, "__write_profiling");
    if (monitorCleanup != 0) {
        voidFuncCallFP rout = (voidFuncCallFP) monitorCleanup;
        (*rout)();
    } else {
        fprintf(stderr, "no __write_profiling symbol located\n");
        /* try _mcleanup() as fallback, but it can only be called once */
        static bool didCleanup = false;
        if (didCleanup == false) { /* never called */
            monitorCleanup = dlsym(dlHandle, "_mcleanup");
            if (monitorCleanup != 0) {
                voidFuncCallFP rout = (voidFuncCallFP) monitorCleanup;
                didCleanup = true; /* don't call this again */
                (*rout)();
            } else {
                fprintf(stderr, "no _mcleanup symbol located\n");
            }
        }
    }

    void *gcovFlush = dlsym(dlHandle, "__gcov_flush");
    if (gcovFlush != 0) {
        voidFuncCallFP rout = (voidFuncCallFP) gcovFlush;
        (*rout)();
    } else {
        fprintf(stderr, "no __gcov_flush symbol located\n");
    }
    dlclose(dlHandle);
    switch (forceExit) {
    case 1:
        _exit(1);
        break;
    case 2:
        exit(1);
        break;
    default:
        break;
    }
}

/*! \brief Flush profiling buffers.
 */
#define FLUSH_PROFILING_BUFFERS(forceExit) flushProfilingBuffers(forceExit)

#else
// Win32 doesn't support profiling
// The _exit() semantics might be more closely achieved by TerminateProcess()
#define FLUSH_PROFILING_BUFFERS(forceExit) do { \
    if (forceExit == 1) TerminateProcess(GetCurrentProcess(), 1); \
    if (forceExit == 2) ExitProcess(1); \
} while (0)

#endif

#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
