/* Copyright (C) 2011 - 2020 FARGOS Development, LLC */
#ifndef _MAPPED_FILE_H_
//! \cond
#define _MAPPED_FILE_H_ "$Id"
//! \endcond

#ifdef _WIN32
#include <winsock2.h>
#include <windows.h>
#endif

#include <stdint.h>

#ifdef __cplusplus

/** \addtogroup FARGOSio 
 *
 * @{
 */

/*! \addtogroup FARGOSmappedFiles Memory-Mapped File Functions
 *
 * \brief Utility functions to create and open memory-mapped files.
 *
 * @{
 */

extern "C" {
#endif

#define FILENAME_WITH_DATE 1    /*!< Filename should be date-qualified */
#define FILENAME_WITH_PID 2     /*!< Filename should be pid-qualified */
#define FILENAME_WITH_DATE_AND_PID 3 /*!< Filename should be date- and pid-qualified */
#define FILENAME_WITH_TIME 4     /*!< Filename should be time-qualified */
#define FILENAME_WITH_DATE_AND_TIME 5 /*!< Filename should be date- and time-qualified */
#define FILENAME_WITH_HOST 8 /*!< Filename should be hostname-qualified */
#define FILENAME_ADD_DIR_PREFIX 16 /*!< Unqualified filename should be prepended with default directory */
#define FILENAME_WITH_EVERYTHING ~0 /*!< Filename should be date-, time- and pid-qualified */

/*! \brief Create a filename and optionally qualify with the current date
 * and process Id.
 *
 * \param fileName points to a buffer into which the constructed file name
 * will be stored.
 * \param fileNameLen specifies the length of the \c fileName buffer.
 * \param componentName specifies the name of the component creating the file.
 * \param suffix specifies a suffix for the constructed file name,
 * often something like ".log" or ".txt".
 * \param dirName specifies the directory in which the file should reside.
 * \param flags is a bit mask that specifies a combination of attributes
 * that should be included in the filename.  FILENAME_WITH_EVERYTHING
 * includes all components whereas FILENAME_WITH_DATE would only include
 * the current date.
 *
 * \return Length of constructed file name that was stored in \c fileName.
 */
uint_fast32_t makeQualifiedFileName(char *fileName, uint_fast32_t fileNameLen,
    const char *componentName, const char *suffix,
    const char *dirName, uint_fast32_t flags);

/* TODO:  add argument for options like:
 * always zero out file?
 * zero out to full length on initial create?
 */

#define MAP_FILE_READONLY 0     /*!< Mapped file is read-only */
#define MAP_FILE_READWRITE 1    /*!< Mapped file is read/write */
#define MAP_FILE_RESET 2        /*!< Reset mapped file */
#define MAP_FILE_INIT 3         /*!< Initialize mapped file */

/*! \brief Write out zero-filled blocks to a file.
 *
 * \param descriptor references a file already opened for write.
 * \param fileLength specifies the number of bytes to be written.
 */
#ifndef _WIN32
size_t zeroFileContents(int descriptor, size_t fileLength);
#else
size_t zeroFileContents(HANDLE handle, size_t fileLength);
#endif

/*! \brief Create or open a mapped file with the specified file name.
 *
 * \param segment is pointer to variable into which will be stored
 * the base of the mapped segment.  A value of 0 (which would represent
 * a null pointer) will be returned on any error and is a portable
 * way to detect the condition.
 * \param fileName specifies the name of the file to be opened/created.
 * The file must exist unless the MAP_FILE_RESET bit is set in the \c doInit
 * argument.
 * \param segmentLenPtr optionally points to a variable into which the
 * segment length will be stored.  If a null pointer or the existing
 * value in the pointed-to variable is zero, the current length of the
 * file will be used to determine the size of the segment.  A value must be
 * provided if a new file is being created.
 * \param doInit indicates how the segment should be prepared.  MAP_FILE_INIT
 * is used to create new segments or reinitialize an existing one.
 * MAP_FILE_READONLY is used to obtain read-only access to an existing segment.
 * MAP_FILE_READWRITE is used to obtain writeable-access to an existing
 * segment while preserving its existing content.
 *
 * \return Descriptor of opened file.  If segmentLenPtr was not null,
 * the actual size of the segment will be saved in the indicated location.
 * \retval -1 (POSIX) file could not be opened
 * \retval INVALID_HANDLE_VALUE (Windows) file could not be opened
 */
#ifndef _WIN32
int createMappedFile(unsigned char **segment, const char *fileName,
    size_t *segmentLenPtr, uint_fast32_t doInit);
#else
HANDLE createMappedFile(unsigned char **segment, const char *fileName,
    size_t *segmentLenPtr, uint_fast32_t doInit);
#endif


/*! \brief Expand a path with symbolic links into its true name.
 * \param resultPath points to the output buffer into which the result
 * will be stored.  Unlink realpath(), the result is not forced
 * to be an absolute path name.
 * \param bfrLen indicates the size of the provided output buffer.
 * \param priorPath points to the path to be expanded; it does not have to be
 * null-terminated.
 * \param priorLen indicates the length of the path to be expanded.
 * \param mustExist is a Boolean that indicates if all path components
 * must reference directory entries that exist.
 * \return the length of the expanded string placed into \c resultPath is
 * returned.
 * \retval -1 is returned if \c mustExist is true and a path element does not exist.
 * The last path component present in \c resultPath does not exist.
 */
int expand_symlinks(char *resultPath, const uint_fast32_t bfrLen,
    const char *priorPath, uint_fast32_t priorLen, bool mustExist);

/*! \brief Search for file under colon-separated path
 *
 * \param resultFile points to a buffer into which the name of the
 * located file will be stored
 * \param resultBfrLen specifies the size of the provided buffer
 * \param desiredFile specifies the base element to be searched
 * \param searchPath specifies the colon-separated path elements to be
 * searched
 * \param suffix specifies an optional suffix to be added to the end
 * of the desiredFile element.
 *
 * \return -1 is returned if the file cannot be found; otherwise
 * the length of the generated path name that first matched the
 * desired file is returned.  The complete file name will be found in
 * the \c resultFile buffer.
 */
int findFileInPath(char *resultFile, uint_fast32_t resultBfrLen,
        const char *desiredFile, const char *searchPath, const char *suffix="");


#ifdef __cplusplus
}; /* close extern "C" scope */
#endif

/** @} */

/** @} */
#endif

/* vim: set expandtab shiftwidth=4 tabstop=4: */
