#ifndef _MAPPED_VARS_H_
//! \cond
#define _MAPPED_VARS_H_ "$Id: mapped_vars.h 465 2020-07-24 22:59:56Z fargos $"
//! \endcond

/* Copyright (C) 2011 - 2019 FARGOS Development, LLC */

#include <utils/io/mapped_file.h>

/** \addtogroup FARGOSio
 * @{
 */

/*! \defgroup FARGOSmappedVar Memory-mapped Arrays of Variables
 *
 * \brief Creates memory-mapped arrays of variables.
 *
 * @{
 */

#ifdef __cplusplus
extern "C" {
#endif

/*! \brief Data structure to reference a variable record within a
 * memory-mapped segment.
 */    
struct MappedVarRecord {
    unsigned char   *address;       /*!< address of value */
    unsigned char   *outputAddress; /*!< address of output area for formatted value */
    const char      *varName;   /*!< variable name */
    const char      *formatPattern; /*!< pattern to format variable for data */
    uint32_t        len;        /* NOT CURRENTLY USED */
    union {
        uint64_t    u64;
        int64_t     i64;
        int32_t     i32;
        uint32_t    u32;
        float       f;
        double      d;
        unsigned char *cp;
    } val;              /*!< default storage for most types */
};

/*! \brief Create/reload an array of MappedVarRecords that is mapped by
 * a memory mapped file.
 *
 * \param numVars indicates the number of elements in the \c vars array.
 * \param vars is an prepared array of MappedVarRecords (discussed below)
 * and will be filled in with additional attributes upon successful completion.
 * \param segment is the address of a pointer to the shared memory segment.
 * Normally set to null before entry and filled in with the value upon return.
 * \param segmentLenPtr points to the length of the segment.  Normally
 * the value pointed to specifies the desired segment length.  Filled in
 * with the actual segment length upon return.
 * \param doInit is a flag from \c createMappedFile().
 * \param fileName specifies the file name to be created or mapped into memory.
 *
 * \retval -1 indicates the call was unsuccessful.
 * \retval 0 indicates the call was successful, but the existing variables
 * in the mapped segment may not have matched those defined in
 * the \c vars array.  This will always be the case when using unlabeled
 * variables.
 * \retval 1 indicates the call was successful and the variables in
 * the \c vars array matched the contents of the segment.
 * 
 * Usage Notes:
 * There are two basic choices to be made:  whether a pure binary format is
 * to be used or if a plain text format is to be used.  If plain text is used,
 * a second choice is made regarding whether variable names will be used
 * in an attribute/value format or just values.
 *
 * Under normal usage patterns, the address and outputAddress members will
 * be null upon entry.  If formatPattern is null, raw binary format will be
 * used.  If varName is null, no variable name will be provided in text
 * mode.
 *
 */
int createMappedVars(const uint_fast32_t numVars, struct MappedVarRecord vars[],
    unsigned char **segment, size_t *segmentLenPtr, const uint_fast32_t doInit,
    const char *fileName);

/*! \brief Convenience macro to define an initialization record for
 * a MappedVarRecord structure.
 * 
 * \param name points to the variable name
 * \param fmt points to the desired output format
 * \param len specifies the length of the variable data.
 */
#define INIT_MAPPED_VAR(name, fmt, len) { 0, 0, name, fmt, len, 0}

/*! \brief Convenience macro to define an initialization record
 * for a MappedVarRecord representing an integer variable using
 * text output.  The output format used is "=%d", yielding output
 * lines like "varName=NNNN".
 * \param name points to the variable name
 */
#define INIT_MAPPED_INT_VAR(name) INIT_MAPPED_VAR(name, "=%d", sizeof(int))

/* Would be a reinterpret_cast<> in C++, but these are intended to be usable
 * by C source
 */
/*! \brief Get the address of a MappedVarRecords value
 * \param t specifies the type of the data
 * \param tbl specifies the MappedVarRecords array of interest
 * \param id specifies the subscript within the \c tbl.
 */
#define MAPPED_TYPE_VAR(t, tbl, id) ((t *)(tbl[id].address))

/*! \brief Get the address of a MappedVarRecords value representing an integer.
 * \param tbl specifies the MappedVarRecords array of interest
 * \param id specifies the subscript within the \c tbl.
 */
#define MAPPED_INT_VAR(tbl, id) MAPPED_TYPE_VAR(int32_t, tbl,id)

/*! \brief Update (format) the output value of a MappedVarRecord element.
 * \param t specifies the type of the data
 * \param tbl specifies the MappedVarRecord array of interest
 * \param id specifies the subscript within the \c tbl.
 */
#define UPDATE_MAPPED_TYPE_VAR(t, tbl, id) { int _l = sprintf((char *) tbl[id].outputAddress, tbl[id].formatPattern, *MAPPED_TYPE_VAR(t, tbl, id)); tbl[id].outputAddress[_l] = ' '; }

/*! \brief Reload the value represented by the formatted text into the
 * native format held by a MappedVarRecord.
 * \param t specifies the type of the data
 * \param tbl specifies the MappedVarRecord array of interest
 * \param id specifies the subscript within the \c tbl.
 */
#define RELOAD_MAPPED_TYPE_VAR(t, tbl, id) { int _l = sscanf((char *) tbl[id].outputAddress, tbl[id].formatPattern, MAPPED_TYPE_VAR(t, tbl, id)); }

#define UPDATE_MAPPED_INT_VAR(tbl, id) { uint32_t workBfr[6]; uint32_t _l; uint64_t _v = *MAPPED_INT_VAR(tbl, id); char *_r = int_to_ascii(workBfr, sizeof(workBfr), _v, &_l); int _i = 0; while (tbl[id].formatPattern[_i] != '%') { tbl[id].outputAddress[_i] = tbl[id].formatPattern[_i]; _i+=1; } memcpy(tbl[id].outputAddress+_i, _r, _l); }


#ifdef __cplusplus
}; /* close extern "C" scope */

#if 0
// C++-specific additions
template <typename T> struct MappedVarRecordOfType {
};
#endif

#endif

/** @} */

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
