#ifndef _TICK_TIME_H_
//! \cond
#define _TICK_TIME_H_ "$Id: tick_time.h 421 2020-07-14 17:11:44Z fargos $"
//! \endcond

#include <utils/base/compiler_hints.h>

#include <stdint.h>
#if defined(_MSC_VER) || (__MINGW32__ == 1) /* Microsoft Visual Studio */
#include <intrin.h>
#endif

#ifndef _WIN32
#include <sys/time.h>
#else
#ifndef CLOCK_REALTIME
#if (_MSC_VER >= 1900) || (__MINGW32__ == 1)
#include <time.h>
#else
struct timespec {
    time_t  tv_sec;
    long    tv_nsec;
};
#endif

#define CLOCK_REALTIME 1
#endif
#endif

/*! \file
 * \brief FARGOS CPU tick routines.
 */

/** \addtogroup FARGOStime
 * @{
 */


/*! \brief Set to 1 if interpolated time should be used; 0 if native
 * clock_gettime() call should always be used.
 */
#define USE_INTERPOLATED_TIME 1

#ifdef __cplusplus
extern "C" {
#endif


#ifdef _MSC_VER
#define get_CPU_ticks() __rdtsc()
#else /* not Microsoft Visual Studio */

#if defined(__x86_64__) || defined(__i386__)

/*  this is an x86-64-specific implementation for g++. */
#define rdtscll(val) do { \
                 uint32_t __a,__d; \
                 __asm__ __volatile__("rdtsc" : "=a" (__a), "=d" (__d)); \
                 (val) = ((uint64_t)__a) | (((uint64_t)__d)<<32); \
} while(0)
// end x86_64 or i386
#elif defined(__sparc__)
#    ifndef __arch64__
inline uint64_t readSPARCtick()
{
    unsigned long long result;

    result = gethrtime();
    return (result);
}
#    else
// 64-bit SPARC v9
inline uint64_t readSPARCtick()
{
    unsigned long long result;

    __asm__ __volatile__(
        "rd %%tick, %0"
        : "=r"(result));
    return (result);
}
#    endif // __arch64__
// end  __sparc__
#elif defined(__s390__)  // Covers both s390 and s390x.
inline uint64_t readS390timestamp()
{
  // Return the CPU clock.
  uint64_t tsc;
  asm("stck %0" : "=Q"(tsc) : : "cc");
  return (tsc);
}
#elif defined(__aarch64__) // 64-bit ARM
inline uint64_t readAARCH64timer()
{
  // System timer of ARMv8 runs at a different frequency than the CPU's.
  // The frequency is fixed, typically in the range 1-50MHz.  It can be
  // read at CNTFRQ special register.  We assume the OS has set up
  // the virtual timer properly.
  int64_t virtual_timer_value;
  asm volatile("mrs %0, cntvct_el0" : "=r"(virtual_timer_value));
  return (virtual_timer_value);
}
#elif defined(__ARM_ARCH)
  // V6 is the earliest arch that has a standard cyclecount
  // Native Client validator doesn't allow MRC instructions.
#if (__ARM_ARCH >= 6)
inline uint64_t readARMperfCounter()
{
  uint32_t pmccntr;
  uint32_t pmuseren;
  uint32_t pmcntenset;
  // Read the user mode perf monitor counter access permissions.
  asm volatile("mrc p15, 0, %0, c9, c14, 0" : "=r"(pmuseren));
  if (pmuseren & 1) {  // Allows reading perfmon counters for user mode code.
    asm volatile("mrc p15, 0, %0, c9, c12, 1" : "=r"(pmcntenset));
    if (pmcntenset & 0x80000000ul) {  // Is it counting?
      asm volatile("mrc p15, 0, %0, c9, c13, 0" : "=r"(pmccntr));
      // The counter is set up to count every 64th cycle
      return static_cast<int64_t>(pmccntr) * 64;  // Should optimize to << 6
    }
  }
  return (0); // not supported
}
#endif /* ARM_ARCH >= 6 */
#elif defined(__linux__)
// final fall back for Linux
#include <linux/perf_event.h>
#include <asm/unistd.h>
#include <sys/ioctl.h>
// used as final fall back on Linux
extern int initPerfEvent();
extern uint64_t getCPUcyclesPerfEvent();
extern int perfEvent_FD;
#endif

extern uint64_t get_CPU_ticks() OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");

/*! \brief Return value of CPU tick register.
 *
 */
inline uint64_t get_CPU_ticks()
{
    uint64_t result;

#if defined(__i386__) || defined(__x86_64__)
    rdtscll(result);
#elif defined(__sparc__)
    result = readSPARCtick();
#elif defined(_WIN32)
    result = rdtsc();
#elif defined(__s390__)
    result = readS390timestamp();
#elif defined(__aarch64__) // 64-bit ARM
    result = readAARCH64timer();
#elif defined(__ARM_ARCH_7A__) || defined(__ARM_ARCH_8A)
    result = readARMperfCounter();
#elif defined(__linux__)
#    warning "CPU type not yet supported by OMEgetCPUticks()\n";
    if (perfEvent_FD == -1) { // not yet initialized
        int rc = initPerfEvent();
        if (rc == -1) {
            result = 0;
            return (result);
        }
    }
    result = getCPUcyclesPerfEvent();
#if 0
    uint32_t r = 0;
    asm volatile("mrc p15, 0, %0, c9, c13, 0" : "=r"(r));
    result = r;
#endif
#else
#    error "CPU type not supported by get_CPU_ticks()\n";
    result = 0;
#endif
    return (result);
}

#endif /* not Microsoft Visual Studio */

/*! \brief Manually set CPU tick rate per nanoseconds */
void set_tick_rate(double ticksPerNanosecond);

/*! \brief Set maximum number of nanoseconds between forced
 * retrieval of time from hardware clock when clock_gettime_via_tick()
 * is called with force=false.  Defaults to 1 second.
 * NOTE: setting to 0 will have the effect of preventing any possible
 * use of the interpolated clock.
 */
void set_max_clock_sample_interval(uint64_t nanoseconds);

/*! \brief Normal initialization routine to determine tick rate
 * of underlying hardware.
 *
 * \param setGlobalRate is a Boolean flag which indicates if
 * a call to set_tick_rate() should be performed to set the
 * discovered rate.
 *
 * \return The number of ticks per nanosecond times 100.
 *
 * \note This routine takes a few seconds to execute; it is suitable
 * for one-time initialization in a long-running program.
 * For short-lived programs, like utilities used in shell scripts,
 * the get_clock_rate_estimate() routine is a better choice.
 */
uint64_t profile_clock_rate(bool setGlobalRate);

#if __linux__ == 1
/*! \brief Linux-specific routine to determine tick rate
 * of underlying hardware as reported by the kernel via /proc/cpuinfo.
 *
 * \param setGlobalRate is a Boolean flag which indicates if
 * a call to set_tick_rate() should be performed to set the
 * discovered rate.
 *
 * \return The number of ticks per nanosecond times 100.
 *
 * \note This routine determines the tick rate quickly in constant time,
 * but is not as precise as profile_clock_rate().  Its implementation is
 * also Linux-specific.  It is intended for use in short-lived utilities,
 * such as command line utilities invoked by shell scripts.
 */
uint64_t get_clock_rate_estimate(bool setGlobalRate);
#endif

/*! \brief Defines the interface for clock_gettime_via_tick()
 * or equivalent.
 *
 * \param result points to the storage into which the
 * current time will be copied.
 * \param force indicates if interpolation is disabled.
 * 
 * This is not an ideal interface, but it inherits its signature
 * from the POSIX clock_gettime() function.
 *
 * \retval 0 indicates success
 * \retval -1 indicates failure
 */
typedef int (*Clock_GetTime_InterfaceFP)(struct timespec *result, bool force);

/*! \brief Get time with nanosecond precision using interpolation
 * of elapsed CPU ticks.
 *
 * \param result is a pointer to a timespec structure which will
 * be filled in with the current time.
 * \param force is a Boolean which indicates if the use of
 * clock_gettime() should be used rather than attempt to calculate
 * a time based on elapsed CPU ticks.
 *
 * \note If nanosecond precision is absolutely required, the \a force
 * flag should be set to true.  If microsecond precision is tolerable,
 * the \a force flag can be set to false to avoid the overhead of always
 * calling clock_gettime().  If too many CPU ticks have occurred between calls
 * to clock_gettime(), the current time will be obtained from clock_gettime().
 */
int interpolated_clock_gettime(struct timespec *result, bool force);

/*! \brief Alternate of interpolated_clock_gettime() that can use an
 * external RDTSC counter.
 */
int simulated_clock_gettime(struct timespec *result, bool force);

/*! \brief Sets the storage location from which the current
 * time in nanoseconds will be read.
 *
 * \param rdtsc_pointer points at the storage location to be used
 *
 * As a convenience, this routine automatically selects the use of
 * simulated_clock_gettime().   If an alternate clock source routine
 * is desired, a subsequent call to set_clock_function() must be made.
 */
void set_external_rdtsc_variable(uint64_t *rdtsc_pointer);

/*! \brief Set clock routine implementation.
 *
 * \param func specifies the function to be used.
 *
 * The default clock source function is interpolated_clock_gettime().
 *
 * \return The previously set clock routine is returned.
 */
Clock_GetTime_InterfaceFP set_clock_function(Clock_GetTime_InterfaceFP func);

extern Clock_GetTime_InterfaceFP clockRoutine_function;

#if USE_INTERPOLATED_TIME == 1
// declare first with attributes
extern int clock_gettime_via_tick(struct timespec *result, bool force) OME_ALWAYS_INLINE OME_ALWAYS_OPTIMIZE("-O3");

inline int clock_gettime_via_tick(struct timespec *result, bool force)
{
    // make indirect call to active clock routine
    return ((*clockRoutine_function)(result, force));
}

#else
/*! \brief Compatibility macro to directly invoke the
 * standard routine clock_gettime().
 */
#define clock_gettime_via_tick(r,f) clock_gettime(CLOCK_REALTIME, r)
#endif


#ifdef __cplusplus
}; // end extern C scope
#endif

/** @} */
#endif
/* vim: set expandtab shiftwidth=4 tabstop=4: */
